import { Box, Button, Input, Switch, Table, Text, useToast } from 'ugnis';
import React, { ChangeEvent, useEffect, useMemo, useState } from 'react';
import { MdFileUpload } from 'react-icons/md';
import { useClient, useQuery, RequestPolicy } from 'urql';
import debounce from 'lodash/debounce';
import { useLape } from 'lape';
import { Loading } from '../../../components';
import { GET_TABLE } from '../../TableOfTables/gql';
import { pathToParams, replace, ROUTES } from '../../../routes';

interface Props {
  id: string;
}

const SEARCH = 'Search';
const UPLOAD = 'Upload';

const EditableTableInternal = ({ id }: Props) => {
  const toast = useToast();
  const client = useClient();
  const [result] = useQuery({
    query: GET_TABLE,
    variables: {
      input: id,
    },
  });

  const table = result?.data?.system?.table;
  const context = useMemo(
    () => ({
      additionalTypenames: [table?.uniqueIdentifier!],
      requestPolicy: 'network-only' as RequestPolicy,
    }),
    [table?.uniqueIdentifier],
  );

  const [search, setSearch] = useState('');
  const onSearchChange = debounce((e: ChangeEvent<HTMLInputElement>) => {
    setSearch(e.target.value);
  }, 300);
  const state = useLape<{
    tableName?: string;
    data?: any[];
    loading: boolean;
    errorCount: number;
    filterErrors: boolean;
    saveStatus: 'saving' | 'saved' | 'error' | null;
  }>({
    tableName: undefined,
    data: undefined,
    loading: true,
    errorCount: 0,
    filterErrors: false,
    saveStatus: null,
  });

  const updateRowSyncBackend = debounce(
    (
      id: string,
      columnName: string,
      value: string | number | boolean,
      originalValue: string | number | boolean,
      hasError: boolean,
    ) => {
      state.saveStatus = 'saving';
      client
        .mutation(
          `
            mutation updateData($input: Update${table!.uniqueIdentifier}Input!){
              response: update${table!.uniqueIdentifier}(input: $input) {
                resp: ${table!.uniqueIdentifier} {
                  ${columnName}
                  _errors {
                    fieldName
                    errorText
                  }
                }
              }
            }
        `,
          {
            input: {
              filter: {
                id: id,
              },
              set: {
                [columnName]: value,
              },
            },
          },
          context,
        )
        .toPromise()
        .then(({ data, error }) => {
          if (error) {
            throw error;
          }
          const rowIndex = state.data?.findIndex((row) => row.id === id);
          if (state.data && rowIndex && rowIndex !== -1) {
            state.data[rowIndex] = {
              ...state.data[rowIndex],
              ...data.response.resp,
            };
          }
          const errorResolved =
            hasError && !data.response.resp?._errors?.lenght;
          if (errorResolved) {
            state.errorCount--;
          }
          state.saveStatus = 'saved';
        })
        .catch((error) => {
          const rowIndex = state.data?.findIndex((row) => row.id === id);
          toast({
            title: 'Error has occured',
            description: error.message,
            status: 'error',
          });
          state.saveStatus = 'error';
          if (state.data && rowIndex && rowIndex !== -1) {
            state.data[rowIndex][columnName] = originalValue;
          }
        })
        .finally(() => {
          state.loading = false;
        });
    },
    300,
  );

  const updateRow = (
    row: {
      id: string;
      _error: any[] | null;
      [key: string]: any;
    },
    columnName: string,
  ) => {
    const originalValue = row[columnName];
    return (e: ChangeEvent<HTMLInputElement>) => {
      row[columnName] = e.target.value;
      updateRowSyncBackend(
        row.id,
        columnName,
        e.target.value,
        originalValue,
        !!row._errors?.length,
      );
    };
  };

  const updateSwitch = (
    row: {
      id: string;
      _error: any[] | null;
      [key: string]: any;
    },
    columnName: string,
  ) => {
    const originalValue = row[columnName];
    return () => {
      row[columnName] = !row[columnName];
      updateRowSyncBackend(
        row.id,
        columnName,
        row[columnName],
        originalValue,
        !!row._errors?.length,
      );
    };
  };

  useEffect(() => {
    if (table?.uniqueIdentifier) {
      client
        .query(
          `
              query aggregateErrors($filter: ${table.uniqueIdentifier}Filter!) {
                  response: aggregate${table.uniqueIdentifier}(filter: $filter) {
                      count
                  }
              }
          `,
          {
            filter: { has: '_errors' },
          },
        )
        .toPromise()
        .then(({ data }) => {
          state.errorCount = data.response.count;
        });
    }
  }, [table?.uniqueIdentifier]);

  useEffect(() => {
    if (table?.uniqueIdentifier) {
      client
        .query(
          `
            query findData($filter: ${table.uniqueIdentifier}Filter!){
              response: query${table.uniqueIdentifier}(filter: $filter) {
                ${table.columns.map((col) => col.uniqueIdentifier).join('\n')}
                id
                _errors {
                  fieldName
                  errorText
                }
              }
            }
        `,

          {
            filter: state.filterErrors ? { has: '_errors' } : {},
          },
          context,
        )
        .toPromise()
        .then(({ data }) => {
          if (!data.response?.length) {
            replace(ROUTES.uploadTable, { id });
          }
          state.data = data.response;
        })
        .finally(() => {
          state.loading = false;
        });
    }
  }, [table?.uniqueIdentifier, state.filterErrors]);

  const columns = table?.columns.map((column) => ({
    key: column.uniqueIdentifier,
    name: column.name,
    resizable: true,
    editable: column.type !== I_ColumnType.Boolean,
    editor: ({ row }: any) => (
      <Box pl="0.5rem">
        <Input
          variant="unstyled"
          autoFocus
          size="small"
          value={row[column.uniqueIdentifier]}
          onChange={useMemo(() => updateRow(row, column.uniqueIdentifier), [])}
        />
      </Box>
    ),
    editorOptions: {
      editOnClick: true,
    },
    formatter: ({ row }: any) =>
      column.type === I_ColumnType.Boolean ? (
        <Box display="flex" alignItems="center" height="100%">
          <Switch
            onChange={updateSwitch(row, column.uniqueIdentifier)}
            defaultChecked={row[column.uniqueIdentifier]}
          />
        </Box>
      ) : (
        row[column.uniqueIdentifier]
      ),
    cellClass: (row: { _errors: I_Errors[] | null; id: string }) =>
      row._errors?.length &&
      row._errors.find((err) => err.fieldName === column.uniqueIdentifier)
        ? 'rdg-cell-error'
        : '',
  }));

  if (state.loading || !table || !columns) {
    return <Loading />;
  }

  return (
    <>
      <Box minH="5.375rem" display="flex" alignItems="center" px="0.75rem">
        <Box display="flex" alignItems="baseline">
          <Text variant="h5" mr="1rem">
            {table.name}
          </Text>
          <Text variant="body" mr="1rem" color="neutral.400">
            {state.saveStatus === 'saving' && 'Saving...'}
            {state.saveStatus === 'saved' && 'Saved'}
            {state.saveStatus === 'error' && 'Error - changes were not saved'}
          </Text>
        </Box>
        <Box display="flex" flex={1} justifyContent="flex-end">
          <Box w="12rem" mr="1rem">
            <Input placeholder={SEARCH} size="sm" onChange={onSearchChange} />
          </Box>
          <Button size="sm" variant="outline" rightIcon={<MdFileUpload />}>
            {UPLOAD}
          </Button>
        </Box>
      </Box>
      {!!state.errorCount && (
        <Box
          bg="error.50"
          mx={1}
          mb={1}
          px={1}
          py={0.5}
          borderRadius="8px"
          display="flex"
          justifyContent="space-between"
          alignItems="center"
        >
          {state.errorCount === 1 ? (
            <Text>
              There is{' '}
              <Text fontWeight="bold" display="inline">
                {state.errorCount}
              </Text>{' '}
              error in this table
            </Text>
          ) : (
            <Text>
              There are{' '}
              <Text fontWeight="bold" display="inline">
                {state.errorCount}
              </Text>{' '}
              errors in this table
            </Text>
          )}
          <Box>
            <Switch
              textPosition="left"
              onChange={() => (state.filterErrors = !state.filterErrors)}
            >
              Show only rows with errors
            </Switch>
          </Box>
        </Box>
      )}
      <Box display="flex" flex={1} px={1}>
        {!state.data?.length ? (
          <Text>No data yet</Text>
        ) : (
          <Table
            rows={state.data as any}
            columns={columns}
            rowKeyGetter={(row) => row.id}
          />
        )}
      </Box>
    </>
  );
};

export const EditableTable = () => {
  const { id } = pathToParams(ROUTES.table);

  // Rerender component when id changes, so we wouldn't have to reset all state manually
  return <EditableTableInternal key={id} id={id} />;
};
