import { Upload } from './Upload';
import { ignoreState, useLape } from 'lape';
import { jsonToGraphQLQuery } from 'json-to-graphql-query';
import { SelectHeader } from './SelectHeader';
import { MatchColumns } from './MatchColumns';
import { useEffect } from 'react';
import { navigate, pathToParams, ROUTES } from '../../../routes';
import { IGetTableQuery } from '../../TableOfTables/gql';
import { getTable } from '../api/getTable';
import { Loading } from '../../../components';
import { newTable } from '../api/newTable';
import { ErrorPage } from '../../../components/ErrorPage';
import { CombinedError } from 'urql';
import { generateInitialValues } from '../utils/generateInitialValues';
import { generateOptions } from '../utils/generateOptions';
import { serializeData } from '../utils/serializeData';

enum Type {
  upload,
  selectHeader,
  loadMatchColumns,
  loadMatchColumnsError,
  matchColumns,
  submitMatchColumns,
  submitMatchColumnsError,
}

type StateType =
  | {
      type: Type.upload;
    }
  | {
      type: Type.selectHeader;
      data: string[][];
    }
  | {
      type: Type.loadMatchColumns;
      data: string[][];
      headerValues: string[];
    }
  | {
      type: Type.loadMatchColumnsError;
      data: string[][];
      headerValues: string[];
      error: string;
    }
  | {
      type: Type.matchColumns;
      data: string[][];
      headerValues: string[];
      table: IGetTableQuery;
      values: object;
    }
  | {
      type: Type.submitMatchColumns;
      data: string[][];
      headerValues: string[];
      table: IGetTableQuery;
      values: object;
    }
  | {
      type: Type.submitMatchColumnsError;
      data: string[][];
      headerValues: string[];
      table: IGetTableQuery;
      values: object;
      error: string;
    };

type State = {
  step: StateType;
};

export const UploadFlow = () => {
  const { id } = pathToParams(ROUTES.table);
  const state = useLape<State>({
    step: { type: Type.upload },
  });

  useEffect(() => {
    const fetchApi = async () => {
      if (state.step.type === Type.loadMatchColumns) {
        try {
          const table = await getTable(id);
          const options = generateOptions(table);
          const initialValues = generateInitialValues(
            state.step.headerValues,
            options,
          );

          state.step = ignoreState({
            type: Type.matchColumns,
            data: state.step.data,
            headerValues: state.step.headerValues,
            table: table,
            values: initialValues,
          });
        } catch (e) {
          state.step = ignoreState({
            type: Type.loadMatchColumnsError,
            data: state.step.data,
            headerValues: state.step.headerValues,
            error: ((e as CombinedError).message || e) as string,
          });
        }
      }
    };
    fetchApi();
  }, [state.step.type]);

  useEffect(() => {
    const fetchApi = async () => {
      if (state.step.type === Type.submitMatchColumns) {
        const matchedColumnsArr = state.step.headerValues.map(
          (headerValue) => (state.step as any).values[headerValue],
        );
        const mutation = jsonToGraphQLQuery(
          {
            mutation: {
              [`new${state.step.table.get_Table!.uniqueIdentifier}`]: {
                __args: {
                  input: serializeData(state.step.data, matchedColumnsArr),
                },
                id: true,
              },
            },
          },
          { pretty: true },
        );
        try {
          await newTable(mutation);
          navigate(ROUTES.table, { id });
        } catch (e) {
          state.step = ignoreState({
            type: Type.submitMatchColumnsError,
            data: state.step.data,
            headerValues: state.step.headerValues,
            values: state.step.values,
            table: state.step.table,
            error: ((e as CombinedError).message || e) as string,
          });
        }
      }
    };

    fetchApi();
  }, [state.step.type]);

  switch (state.step.type) {
    case Type.upload:
      return (
        <Upload
          onContinue={(data) => {
            state.step = ignoreState({ type: Type.selectHeader, data });
          }}
        />
      );
    case Type.selectHeader:
      return (
        <SelectHeader
          data={state.step.data}
          onContinue={(headerValues: string[], data: string[][]) => {
            state.step = ignoreState({
              type: Type.loadMatchColumns,
              headerValues,
              data,
            });
          }}
          onCancel={() => {
            state.step = ignoreState({ type: Type.upload });
          }}
        />
      );
    case Type.matchColumns:
    case Type.submitMatchColumns:
    case Type.submitMatchColumnsError:
      return (
        <MatchColumns
          headerValues={state.step.headerValues}
          onCancel={() => {
            state.step = ignoreState({ type: Type.upload });
          }}
          values={state.step.values}
          table={state.step.table}
          loading={state.step.type === Type.submitMatchColumns}
          error={
            state.step.type === Type.submitMatchColumnsError
              ? state.step.error
              : undefined
          }
          onContinue={(values) => {
            if (
              state.step.type === Type.matchColumns ||
              state.step.type === Type.submitMatchColumnsError
            ) {
              state.step = ignoreState({
                type: Type.submitMatchColumns,
                data: state.step.data,
                headerValues: state.step.headerValues,
                table: state.step.table,
                values,
              });
            }
          }}
        />
      );
    case Type.loadMatchColumns:
      return <Loading />;

    case Type.loadMatchColumnsError:
      return (
        <ErrorPage
          error={state.step.error}
          onContinue={() => {
            if (state.step.type === Type.loadMatchColumnsError) {
              state.step = ignoreState({
                type: Type.loadMatchColumns,
                data: state.step.data,
                headerValues: state.step.headerValues,
              });
            }
          }}
        />
      );
    default:
      return null;
  }
};
